/*===========================================================================
 * Licensed Materials - Property of IBM
 * "Restricted Materials of IBM"
 * 
 * IBM SDK, Java(tm) Technology Edition, v6
 * (C) Copyright IBM Corp. 2013, 2014. All Rights Reserved
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *===========================================================================
 */
/*
 * %W% %E%
 *
 * Copyright (c) 2006, Oracle and/or its affiliates. All rights reserved.
 * ORACLE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

package java.io;

import java.util.concurrent.atomic.AtomicInteger;                               //IBM-perf_FileIO

/**
 * Instances of the file descriptor class serve as an opaque handle
 * to the underlying machine-specific structure representing an open
 * file, an open socket, or another source or sink of bytes. The
 * main practical use for a file descriptor is to create a
 * <code>FileInputStream</code> or <code>FileOutputStream</code> to
 * contain it.
 * <p>
 * Applications should not create their own file descriptors.
 *
 * @author  Pavani Diwanji
 * @version %I%, %G%
 * @see	    java.io.FileInputStream
 * @see	    java.io.FileOutputStream
 * @since   JDK1.0
 */
public final class FileDescriptor {
 
    private int fd;
    /*JSE-598 starts*/                                                          //IBM-net_bringup
    private int fd2;                                                            //IBM-net_bringup
    private int boundfd;                                                        //IBM-net_bringup
    private int fdinuse;                                                        //IBM-net_bringup
    /*JSE-598 ends*/                                                            //IBM-net_bringup
    private long handle;
	                                                                        //IBM-perf_FileIO
    /**                                                                         //IBM-perf_FileIO
     * A counter for tracking the FIS/FOS/RAF instances that                    //IBM-perf_FileIO
     * use this FileDescriptor. The FIS/FOS.finalize() will not release         //IBM-perf_FileIO
     * the FileDescriptor if it is still under user by a stream by default.     //IBM-perf_FileIO
     * When com.ibm.streams.CloseFDWithStream is set to false,                  //IBM-perf_FileIO
     * Filedescriptors are closed by the last object using them.                //IBM-perf_FileIO
     */                                                                         //IBM-perf_FileIO
    private AtomicInteger useCount;                                             //IBM-perf_FileIO


    /**
     * Constructs an (invalid) FileDescriptor
     * object.
     */
    public /**/ FileDescriptor() {
	fd = -1;
        handle = -1;
	/* JSE-598 starts */                                                    //IBM-net_bringup
	fd2 = -1;                                                               //IBM-net_bringup
	boundfd = -1;                                                           //IBM-net_bringup
	fdinuse = -1;                                                           //IBM-net_bringup
    	/*JSE-598 ends*/                                                    //IBM-perf_FileIO
	useCount = new AtomicInteger();                                         //IBM-perf_FileIO
    }

    private /* */ FileDescriptor(int fd) {
	this.fd = fd;
        handle = -1;
	/* JSE-598 starts */                                                    //IBM-net_bringup
	fd2 = -1;                                                               //IBM-net_bringup
	boundfd = -1;                                                           //IBM-net_bringup
	fdinuse = -1;                                                           //IBM-net_bringup
    	/*JSE-598 ends*/                                                    //IBM-perf_FileIO
	useCount = new AtomicInteger();                                         //IBM-perf_FileIO
    }

    static {
        initIDs();
    }

    int read() throws IOException {                                             //IBM-perf_FileIO
        return read(handle);                                                    //IBM-perf_FileIO
    }                                                                           //IBM-perf_FileIO
                                                                                //IBM-perf_FileIO
    private native int read(long handle) throws IOException;                    //IBM-perf_FileIO
                                                                                //IBM-perf_FileIO
    /**
     * A handle to the standard input stream. Usually, this file
     * descriptor is not used directly, but rather via the input stream
     * known as <code>System.in</code>.
     *
     * @see     java.lang.System#in
     */
    public static final FileDescriptor in = standardStream(0);

    /**
     * A handle to the standard output stream. Usually, this file
     * descriptor is not used directly, but rather via the output stream
     * known as <code>System.out</code>.
     * @see     java.lang.System#out
     */
    public static final FileDescriptor out = standardStream(1);

    /**
     * A handle to the standard error stream. Usually, this file
     * descriptor is not used directly, but rather via the output stream
     * known as <code>System.err</code>.
     *
     * @see     java.lang.System#err
     */
    public static final FileDescriptor err = standardStream(2);

    /**
     * Tests if this file descriptor object is valid.
     *
     * @return  <code>true</code> if the file descriptor object represents a
     *          valid, open file, socket, or other active I/O connection;
     *          <code>false</code> otherwise.
     */
    public boolean valid() {
	return ((handle != -1) || (fd != -1) || (fd2 != -1));                   //IBM-net_bringup
    }

    /**
     * Force all system buffers to synchronize with the underlying
     * device.  This method returns after all modified data and
     * attributes of this FileDescriptor have been written to the
     * relevant device(s).  In particular, if this FileDescriptor
     * refers to a physical storage medium, such as a file in a file
     * system, sync will not return until all in-memory modified copies
     * of buffers associated with this FileDesecriptor have been
     * written to the physical medium.
     *
     * sync is meant to be used by code that requires physical
     * storage (such as a file) to be in a known state  For
     * example, a class that provided a simple transaction facility
     * might use sync to ensure that all changes to a file caused
     * by a given transaction were recorded on a storage medium.
     *
     * sync only affects buffers downstream of this FileDescriptor.  If
     * any in-memory buffering is being done by the application (for
     * example, by a BufferedOutputStream object), those buffers must
     * be flushed into the FileDescriptor (for example, by invoking
     * OutputStream.flush) before that data will be affected by sync.
     *
     * @exception SyncFailedException
     *	      Thrown when the buffers cannot be flushed,
     *	      or because the system cannot guarantee that all the
     *	      buffers have been synchronized with physical media.
     * @since     JDK1.1
     */
    public native void sync() throws SyncFailedException;

   /**                                                                          //IBM-perf_FileIO
    * Increment the reference count to reflect the fact                         //IBM-perf_FileIO
    * that another object is using this FileDescriptor                          //IBM-perf_FileIO
    */                                                                          //IBM-perf_FileIO
    int incrementAndGetUseCount() {                                             //IBM-perf_FileIO
        return useCount.incrementAndGet();                                      //IBM-perf_FileIO
    }                                                                           //IBM-perf_FileIO
                                                                                //IBM-perf_FileIO
    /**                                                                         //IBM-perf_FileIO
     * Decrement the reference count to note that an                            //IBM-perf_FileIO
     * object is no longer using this FileDescriptor                            //IBM-perf_FileIO
     * @return the reference count - when 0, the decementing object             //IBM-perf_FileIO
     * can safely close() the FileDescriptor                                    //IBM-perf_FileIO
     */                                                                         //IBM-perf_FileIO
    int decrementAndGetUseCount() {                                             //IBM-perf_FileIO
        return useCount.decrementAndGet();                                      //IBM-perf_FileIO
    }                                                                           //IBM-perf_FileIO
                                                                                //IBM-perf_FileIO
    /* This routine initializes JNI field offsets for the class */
    private static native void initIDs();

    private static native long set(int d);

    private static FileDescriptor standardStream(int fd) {
        FileDescriptor desc = new FileDescriptor();
        desc.handle = set(fd);
        return desc;
    }


}
//IBM-net_bringup
//IBM-perf_FileIO
